import aiger_bv as BV
import aiger_discrete
from aiger_bv import atom, ite, split_gate, encode_int, lookup
from bidict import bidict

import aiger_gridworld as G


NORTH = '↑'
SOUTH = '↓'
EAST = '→'
WEST = '←'


ACTIONS = bidict({
    #      y   x
    '←': 0b00_01,
    '→': 0b00_10,
    '↓': 0b01_00,
    '↑': 0b10_00,
})


ACTIONS_C = bidict({'←': 0b00, '↑': 0b01, '→': 0b10, '↓': 0b11})
COMPRESSION_MAPPING = {
    ACTIONS_C['→']: ACTIONS['→'],
    ACTIONS_C['←']: ACTIONS['←'],
    ACTIONS_C['↑']: ACTIONS['↑'],
    ACTIONS_C['↓']: ACTIONS['↓'],
}


def chain(n, state_name='x', action='a', start=None, clip=True, can_stay=True):
    if start is None:
        start = n // 2
    else:
        start -= 1

    start = encode_int(n, 1 << start, signed=False)

    x = atom(n, state_name, signed=False)
    a = atom(2, action, signed=False)

    backward, forward = a[0], a[1]
    x2 = ite(forward, x << 1, x >> 1)

    stay = atom(1, 1, signed=False)
    if clip:
        stay = (x2 == 0)
    if can_stay:
        stay |= ~(forward | backward)
    if clip or can_stay:
        x2 = ite(stay, x, x2)

    circ = x2.aigbv['o', {x2.output: state_name}]
    return circ.feedback(
        inputs=[state_name],
        outputs=[state_name],
        initials=[start],
        keep_outputs=True
    )


def gridworld(n, start=(None, None), compressed_inputs=False):
    # Gridworld is 2 synchronized chains.
    circ = chain(n, 'x', 'ax', start[1]) | chain(n, 'y', 'ay', start[0])
    circ <<= split_gate('a', 2, 'ay', 2, 'ax')       # Combine inputs.
    x = BV.uatom(circ.omap['x'].size, 'x')
    y = BV.uatom(circ.omap['y'].size, 'y')
    circ >>= y.concat(x).with_output('state').aigbv  # Combine outputs.

    if compressed_inputs:
        uncompress = lookup(
            2, 4, COMPRESSION_MAPPING, 'a', 'a',
            in_signed=False, out_signed=False
        )
        circ <<= uncompress

    # Wrap using aiger discrete add encoding + valid inputs.
    actions_map = ACTIONS_C if compressed_inputs else ACTIONS
    action_encoding = aiger_discrete.Encoding(
        encode=actions_map.get,
        decode=actions_map.inv.get,
    )
    state_encoding = aiger_discrete.Encoding(
        encode=lambda s: s.yx,
        decode=lambda yx: G.GridState(yx, n),
    )

    func = aiger_discrete.from_aigbv(
        circ,
        input_encodings={'a': action_encoding},
        output_encodings={'state': state_encoding},
    )
    if not compressed_inputs:
        action = BV.uatom(4, 'a')
        is_1hot = (action != 0) & ((action & (action - 1)) == 0)
        func = func.assume(is_1hot)
    return func


__all__ = ['NORTH', 'SOUTH', 'EAST', 'WEST', 'gridworld', 'chain']



import attr


def one_index(val):
    assert val != 0
    count = 0
    while True:
        if (val >> count) & 1:
            break
        count += 1
    return count


@attr.s(auto_attribs=True, frozen=True)
class GridState:
    yx: int
    dim: int

    @property
    def x(self):
        return one_index(self.yx) + 1

    @property
    def y(self):
        return one_index(self.yx >> self.dim) + 1

    def _row(self, row) -> str:
        x, y = self.x, self.y
        for col in range(1, self.dim + 1):
            yield ('▪' if y == row and x == col else '□')

    @property
    def board(self):
        buff = '\n'
        for row in range(self.dim, 0, -1):
            buff += f'{row} '
            buff += ' '.join(self._row(row)) + '\n'
        buff += '  ' + ' '.join(map(str, range(1, self.dim + 1))) + '\n'
        return buff


__all__ = ['GridState']



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiger_gridworld']

package_data = \
{'': ['*']}

install_requires = \
['bidict>=0.22,<0.23',
 'funcy>=1.13,<2.0',
 'py-aiger-bv>=4.0.0,<5.0.0',
 'py-aiger-discrete>=0.1.10,<0.2.0',
 'py-aiger>=6.0.0,<7.0.0']

setup_kwargs = {
    'name': 'py-aiger-gridworld',
    'version': '0.4.3',
    'description': 'Library for modeling gridworlds as AIGER circuits.',
    'long_description': '# py-aiger-gridworld\n\nLibrary for modeling gridworlds as AIGER circuits.\n\n[![Build Status](https://cloud.drone.io/api/badges/mvcisback/py-aiger-gridworld/status.svg)](https://cloud.drone.io/mvcisback/py-aiger-gridworld)\n[![PyPI version](https://badge.fury.io/py/py-aiger-gridworld.svg)](https://badge.fury.io/py/py-aiger-gridworld)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n\n<object data="assets/visualization_example.svg" type="image/svg+xml">\n  <img src="assets/visualization_example.svg" />\n</object>\n\n# Installation\n\nIf you just need to use `aiger_gridworld`, you can just run:\n\n`$ pip install py-aiger-gridwolrd`\n\nFor developers, note that this project uses the\n[poetry](https://poetry.eustace.io/) python package/dependency\nmanagement tool. Please familarize yourself with it and then\nrun:\n\n`$ poetry install`\n',
    'author': 'Marcell Vazquez-Chanlatte',
    'author_email': 'mvc@linux.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
